/******************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * HTTP Daemon
 *
 *  Each HTTP request will create a new Instance object, attach http specific
 *  properties to it and do a URI based routing inside the controllers dir
 *
 *  Each controller will emit via Instance the usual action events (log, error,
 *  output, end) + two additional : flush and header
 *
 *****************************************************************************/

/**
 * Module dependencies
 */
var  httplib = require('http')
    ,url     = require('url')
    ,qs      = require('qs');

/**
 * Httpd
 * @param {number} port
 * @param {function} callback
 * @constructor
 */
function Httpd(port, callback) {

    /**
     * Server port
     * @type {number}
     */
    this.port = port || 1337;

    /**
     * Log file names (will be prefixed by kernel.log)
     * @private
     */
    this._accesslog = 'http.access';
    this._errorlog  = 'http.error';

    /**
     * HTTP Server reference
     * @type {*}
     * @private
     */
    this._server = server.call(this);

    /**
     * List of all running instances
     * @type {{}}
     * @private
     */
    this._running = {};

    /**
     * Number of instances created (since startup)
     * @type {number}
     * @private
     */
    this._count = 0;

    /**
     * Group server status related properties
     * into one public object
     * @type {{}}
     */
    this.status = function() {
        return {

            running : this._running
           ,count   : this._count
           ,port    : this.port

        }
    }.bind(this);

    /**
     * If we got here everything should be ok
     */
    callback();
}

/**
 * Creates the actual http server and defines the logic
 * behind request handling
 * @returns {http}
 */
var server = function() {
    return httplib.createServer(function (req, res) {

        /**
         * Check for post data
         * @type {string}
         */
        var post = '';
        req.on('data', function(chunk) {
            post += chunk.toString();
        });

        /**
         * When the request is completed, create an Instance
         * object and process it
         */
        req.on('end', function() {

            var instance = this.kernel.include(this.kernel.kerneldir + 'instance').new();

            /**
             * Override the instance id to be http specific
             */
            this._count++;
            instance.id = 'http-' + this._count;

            /**
             * Bind http request and response
             */
            instance.request  = req;
            instance.response = res;

            /**
             * get / post shortcuts
             */
            instance.get  = url.parse(instance.request.url, true).query;
            instance.post = {};

            if (post) {
                try {
                    instance.post = JSON.parse(post);
                } catch (e) {
                    instance.post = qs.parse(post);
                }
            }

            /**
             * Internals
             */
            instance._outputbuffer = '';
            instance._httpserver   = this;
            instance._header       = { code : 200, entries : {'Content-type' : 'text/json'}};
            instance._headersent   = false;
            instance._flushed      = false;

            /**
             * Override the Instance events
             */
            instance.echo  = echo;
            instance.error = error;
            instance.log   = log;
            instance.end   = end;

            /**
             * Create additional actions, specific for http communication
             */
            instance.flush  = flush;
            instance.header = header;

            /**
             * Status info
             */
            this._running[instance.id] = {

                 'id'     : instance.id
                ,'req'    : instance.request.url
                ,'start'  : this.kernel.utility.date(instance.starttime, 'Y-m-d h:i:s')
                ,'method' : instance.request.method
                ,'ip'     : instance.request.connection.remoteAddress

            };

            /**
             * Route the request to the appropiate controller
             */
            route.call(instance, this);

        }.bind(this));
    }.bind(this)).listen(this.port);
};

/**
 * Route the request to the appropiate controller based
 * on URI. Only first dir applies, then that controller must
 * handle routing from there.
 * @return
 */
var route = function() {

    var  arr  = url.parse(this.request.url).pathname.split('/')
        ,def  = 'default'
        ,path = def;

    if (arr[1]) {
        path = arr[1];
    }

    /**
     * Try to load the module, if not found
     * will throw
     */
    try {
        this.controller(path).new();
    } catch (err) {
        /**
         * Module not found, this is 404
         */
        if (err.code === 'MODULE_NOT_FOUND') {
            this.end(
                new Error('Invalid route')
            );
        } else {
            this.end(err);
        }
    }
};

/**
 * Mark the instance as ended, flush if needed and remove from
 * the internal skernelboard. Log to the appropiate file depending
 * on error.
 *
 * @return
 */
var end = function(error, data) {

    if (!this._flushed) {
        this.flush(error, data);
    }

    delete this._httpserver._running[this.id];

    var  duration = ((this.kernel.utility.microtime()
                    - this.kernel.utility.microtime(this.starttime)) / 1000) || 0

        ,logmsg   = (this.request.headers['x-forwarded-for'] || this.request.connection.remoteAddress) + ' '
                  +  this.request.method + ' '
                  +  this.request.url + ' '
                  +  duration.toFixed(3) + 's';

    if (error) {
        this.kernel.log(logmsg + ' ' + error, 'http.error', this.starttime);
    } else {
        this.kernel.log(logmsg, 'http.access', this.starttime);
    }
};

/**
 * Flush the buffer (or data) via http.response and end the
 * response. Send the headers if not sent.
 *
 * @param {string|Error} error
 * @param {*} data
 * @return
 */
var flush = function(error, data) {

    if (error instanceof Error) {
        error = error.message;
    }

    if (data && this._outputbuffer) {
        this._outputbuffer += data;
    } else if (data) {
        this._outputbuffer  = data;
    }

    var duration = ((this.kernel.utility.microtime() - this.starttime) / 1000) || 0;

    if (!this._headersent) {
        this.header();
    }

    if (this._header.entries['Content-type'] === 'text/json') {
        this.response.write(
            JSON.stringify({
                'error'    : !!error
                ,'duration' : duration
                ,'response' : (error) ? error : this._outputbuffer
            })
        );
    } else {
        this.response.write(
            (error || this._outputbuffer) + ''
        );
    }

    this.response.end();
    this._flushed = true;
};

/**
 * Sends the http headers and response code
 *
 * @param {number} code
 * @param {object} entries
 * @return
 */
var header = function(code, entries) {

    if (entries) {
        this._header.entries = entries;
    }

    if (code) {
        this._header.code = code;
    }

    this._headersent = true;
    this.response.writeHead(this._header.code, this._header.entries);
};

/**
 * Append data to the outputbuffer
 *
 * @param {string} output
 * @return
 */
var echo = function(output) {
    this._outputbuffer += output;
};

/**
 * Log a message to the http error log
 *
 * @param {string|Error} object
 * @return
 */
var error = function(object) {
    this.kernel.log(object, this._errorlog);
};

/**
 * Log a message to the http log
 *
 * @param {string} message
 * @return
 */
var log = function(message) {
    this.kernel.log(message, this._accesslog);
};

/**
 * Exports
 */
exports = module.exports = Httpd;