/******************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * Spinner
 *
 *  info
 *
 *****************************************************************************/

/**
 * Module dependencies
 */
var fs = require('fs');

/**
 * Spinner
 * @constructor
 */
function Spinner(callback) {

    /**
     * Spinner loop time in ms
     * @type {number}
     * @private
     */
    this._spinspeed = 1000;

    /**
     * List of intervals (as read from the interval dir)
     * @type {}
     * @private
     */
    this._intervals = {};

    /**
     * Running intervals
     * @type {}
     * @private
     */
    this._running = {};

    /**
     * Status object
     * @type {{}}
     * @private
     */
    this.status = function() {
        return {
             speed   : this._spinspeed + 'ms'
            ,lastrun : this.kernel.utility.date(this._lastrun, 'Y-m-d h:i:s')
            ,count   : this._count
            ,running : this._running
            ,pending : this._intervals
        };
    }.bind(this);

    /**
     * Last loop time
     * @type {number}
     * @private
     */
    this._lastrun = 0;

    /**
     * Created instances since startup
     * @type {number}
     * @private
     */
    this._count = 0;

    /**
     * Log prefix (+ additional prefix by kernel.log). All intervals
     * will have output and error logs
     * @private
     */
    this._logprefix = 'interval';

    /**
     * setInterval id
     * @type {number}
     * @private
     */
    this._intv = 0;

    /**
     * Load intervals and start spinning
     */
    intervals.call(this, function(error) {
        if (error) {
            callback(error);
        } else {
            this._intv = setInterval(function() {
                run.call(this);
            }.bind(this), this._spinspeed);
            callback();
        }
    }.bind(this));
}

/**
 * Load intervals from the app dir
 * @return
 */
var intervals = function(callback) {
    var dir = this.kernel.appdir + this.kernel.config._KERNEL_DIR_INTERVALS + '/';
    fs.readdir(dir, function(err, files) {
        if (err) {
            callback(
                new Error('Spinner error : couldn\'t read intervals from : ' + dir)
            );
        } else {

            /**
             * Load each interval
             */
            for (var i = 0, max = files.length; i < max; i++) {

                var  name = files[i].replace('.js','')
                    ,cls  = this.kernel.include(dir + name);

                this._intervals[name] = {
                     active        : cls.active
                    ,maxinstances  : cls.instances
                    ,instances     : 0
                    ,name          : name
                };
            }

            callback();

        }
    }.bind(this));
};

/**
 * Go through the scanned intervals and run them
 * @return
 */
var run = function() {

    /**
     * Last run time
     */
    this._lastrun = this.kernel.utility.microtime();

    /**
     * Loop and run
     */
    for (var name in this._intervals) {
        var interval = this._intervals[name];

        /**
         * If the interval is active and hasn't reached
         * max running instances, run it again
         */
        if (interval.active && interval.instances < interval.maxinstances) {

            var instance = this.kernel.include(this.kernel.kerneldir + 'instance').new();

            /**
             * Override the instance id to be interval specific
             */
            this._count++;
            instance.id = 'interval-' + this._count;

            /**
             * Interval name
             * @private
             */
            instance.name = interval.name;

            /**
             * Override the Instance events
             */
            instance.echo  = echo;
            instance.error = error;
            instance.log   = log;
            instance.end   = end;

            /**
             * Interval log file prefix
             * @private
             */
            instance._log = this._logprefix + '-' + interval.name;

            /**
             * Reference back to the spinner
             * @private
             */
            instance._spinner = this;

            /**
             * Status info
             */
            this._running[instance.id] = {

                 'id'    : instance.id
                ,'name'  : interval.name
                ,'start' : instance.starttime

            };

            /**
             * Increment open instances on the interval class
             */
            this._intervals[interval.name].instances++;

            /**
             * Finally, run the damn thing
             */
            var intvclass = instance.interval(interval.name);

            /**
             * Check for a schedule method
             */
            if (typeof intvclass.schedule === 'function') {
                if (intvclass.schedule()) {
                    intvclass.new();
                } else {
                    /**
                     * If not scheduled we can't construct the interval
                     * so the end method won't get called...so call it
                     * and detach the instance
                     */
                    instance.end();
                }
            } else {
                intvclass.new();
            }
        }
    }
};

/**
 * Echo to the output log
 *
 * @param {string} message
 * @return
 */
var echo = function(message) {
    log.call(this, message);
};

/**
 * Echo to the error log
 *
 * @param {string|Error} message
 * @return
 */
var error = function(message) {
    this.kernel.log(
         message
        ,this._log + '.err'
    );
};

/**
 * Echo to the output log
 *
 * @param {string} message
 * @return
 */
var log = function(message) {
    this.kernel.log(
         message
        ,this._log + '.out'
    );
};

/**
 * Echo to the output or error log and mark the instance
 * as completed (remove it from the status object and free
 * a slot for another instance of this class to be started)
 *
 * @param {string|Error} error
 * @param {string} message
 * @return
 */
var end = function(error, message) {

    if (error) {
        this.error(error);
    } else {
        this.log(message);
    }

    /**
     * Delete from running instances
     */
    delete this._spinner._running[this.id];

    /**
     * Decrement instances for the interval class
     */
    this._spinner._intervals[this.name].instances--;
};

/**
 * Exports
 */
exports = module.exports = Spinner;