/******************************************************************************
 * Lemonade JS
 * https://lemonadejs.com
 *
 * ----
 *
 * Kernel
 *
 *  Holds the include method and references to config and utility. Listens
 *  for log and error events and sends them to their appropriate log locations
 *
 *****************************************************************************/

/**
 * Module dependencies
 */
var  events	= require('events')
    ,util	= require('util')
    ,fs		= require('fs')
    ,url	= require('url')
    ,qs		= require('qs')
    ,os		= require('os')
    ,http	= require('http');

/**
 * Kernel
 *
 * @param {object} config
 * @param {object} boot
 * @param {object} run
 * @param {object} shutdown
 *
 * @constructor
 */
function Kernel(config, boot, run, shutdown) {

    /**
     * Say hello
     */
    console.log();
    console.log('Lemonade ' + this.version + ' (' + this.name + ')');
    console.log();

    /**
     * Dirs
     */
    this.kerneldir  = __dirname + '/';
    this.appdir     = process.cwd() + '/';

    /**
     * Merge config properties
     */
    this.config = this.include(this.kerneldir + 'config').singleton();
    if (config) {
        for (var key in config) {
            this.config[key] = config[key];
        }
    }

    /**
     * Utility methods
     */
    this.utility = this.include(this.kerneldir + 'utility').singleton();

    /**
     * References to status objects or methods from loaded modules
     */
    this._status = {};

    /**
     * Internal counters and references for boot/run/shutdown modules
     * @type {*}
     * @private
     */
    this._boot     = boot;
    this._run      = run;
    this._shutdown = shutdown;
    this._cntboot  = Object.keys(boot).length
    this._cntrun   = Object.keys(run).length;
    this._cntsd    = Object.keys(shutdown).length;

    /**
     * Available loader modules
     * @type {*}
     * @private
     */
    this._modules = fs.readdirSync(this.appdir + this.config._KERNEL_DIR_LOADERS);

    /**
     * Attach listeners and emitters
     * @type {string}
     */
    listen.call(this);
    this.emit('init');
}

/**
 * Add emitter methods to the Kernel
 */
util.inherits(Kernel, events.EventEmitter);

/**
 * Kernel version
 */
Kernel.prototype.version = '0.0.5';
Kernel.prototype.name    = 'Ghiocel';

/**
 * Listen for various events
 * @return
 */
var listen = function() {

    /**
     * Logs a message
     */
    this.on('log', function(message) {
        this.log(message);
    }.bind(this));

    /**
     * Logs an error message
     */
    this.on('error', function(message, shutdown) {
        if (message instanceof Error) {
            this.log(message);
        } else {
            this.log(
                new Error(message)
            );
        }
        if (shutdown) {
            this.emit('shutdown');
        }
    }.bind(this));

    /**
     * Initializes the modules
     */
    this.on('init', function() {

        this.emit('log' , 'Starting '   + this.config.KERNEL_PROJECT_NAME);
        this.emit('log' , 'Logging at ' + this.config.KERNEL_LOG_PATH);
        this.emit('log' , '---');

        for (var key in this._boot) {
            /**
             * Each module will emit "loader-ready" when ready
             */
            attach.call(this, key, this._boot[key]);
        }

        /**
         * Keep checking to see if everything's ready, keep the log updated
         */
        var intv = setInterval(function() {
            if (this._cntboot > 0) {
                this.emit('log' , 'Boot - waiting for loaders : ' + this._cntboot);
            } else if (this._cntrun > 0) {
                this.emit('log' , 'Run - waiting for loaders : ' + this._cntrun);
            }
            if (!this._cntboot && !this._cntrun) {
                clearInterval(intv);
            }
        }.bind(this), 1000);

    }.bind(this));

    /**
     * Each module, when ready, emits "loader-ready"
     */
    this.on('loader-ready', function(name) {
        if (this._cntboot > 0) {
            this.emit('log', 'Init [boot] - ' + name + ' ready');
            if (!--this._cntboot) {
                this.emit('init-boot-ready');
            }
        } else if (this._cntrun > 0) {
            this.emit('log' , 'Init [run] - ' + name + ' ready');
            if (!--this._cntrun) {
                this.emit('log' , 'Init done');
                this.emit('log' , '---');
            }
        }
    }.bind(this));

    /**
     * When all boot modules are ready, load the run modules
     */
    this.on('init-boot-ready', function() {
        for (var key in this._run) {
            attach.call(this, key, this._run[key]);
        }
    }.bind(this));

    /**
     * Something will happen here on shutdown
     */
    this.on('shutdown', function() {
        process.exit(1);
    });
};

/**
 * Kernel adaptors and daemons are linked using the boot script
 * and a loader module. A label is defined in the boot script
 * for the method name which will be used to access the adaptor.
 *
 * var adaptor = kernel.label('switch');
 *
 * @param {string} key
 * @param {*} value
 * @return
 */
var attach = function(key, value) {
    process.nextTick(function() {

        if (typeof value === 'function') {
            /**
             * Value is a function, attach as is
             * @type {*}
             */
            this[key] = value.bind(this);
        } else {
            /**
             * Is value one of the available loader modules?
             */
            for (var i = 0, max = this._modules.length, found = false; i < max; i++) {
                if (value + '.js' === this._modules[i]) {
                    found = true;
                }
            }

            if (found) {
                /**
                 * Attach the loader module
                 * @type {*}
                 */
                this[key] = this.include(
                    this.appdir + this.config._KERNEL_DIR_LOADERS + '/' + value
                ).bind(this);
            } else if (fs.existsSync(this.appdir + value + '.js')) {
                /**
                 * Value is not one of the available modules, treat it
                 * as a string and try to load the path
                 * @type {*}
                 */
                try {
                    this[key] = this.include(this.appdir + value).bind(this);
                } catch (e) {
                    console.log('Couldn\' load : ' + key);
                    process.exit(1);
                }
            }
        }

        /**
         * Execute the newly attached method and, if it has a status
         * object, link it to the kernel status object
         */
        var obj = this[key]();
        if (obj && typeof obj.status !== 'undefined') {
            this._status[key] = obj.status;
        }

    }.bind(this));
};

/**
 * Gather up statuses from the kernel and all loaded modules
 * @return {}
 */
Kernel.prototype.status = function() {

    var  memory = process.memoryUsage()
        ,status = {};

    status.kernel = {

         hostname : os.hostname
        ,memory   : {
             heapUsed  : memory.heapUsed
            ,heapTotal : memory.heapTotal
        }
        ,load : os.loadavg()
        ,node : process.version
        ,lemonade : {
             version : this.version
            ,name    : this.name
        }

    };

    for (var key in this._status) {
        if (typeof this._status[key] === 'function') {
            status[key] = this._status[key]();
        } else {
            status[key] = this._status[key];
        }
    }

    return status;
};

/**
 * Include a class file, overload it to contain instantiating
 * methods, and return it
 *
 * @param {string} path
 * @return {function}
 */
Kernel.prototype.include = function(path) {

    var  Proto   = require(path + '.js')
        ,context = this;

    /**
     * Instantiate a new object
     * @return {object}
     */
    Proto.new = function() {

        /**
         * Overload the object before actually creating it by calling
         * its constructor. Add the kernel and instance references.
         */
        var obj      = Object.create(Proto.prototype || {});
        obj.kernel   = context.kernel || context;
        obj.instance = (context.kernel) ? context : null;

        /**
         * A limit on how many arguments can be used for constructing objects
         * todo : make this better
         */
        Proto.call(obj, arguments[0], arguments[1], arguments[2], arguments[3], arguments[4]);

        return obj;
    };

    /**
     * Singleton based on path and optional key (as param)
     *
     * @param {string} key
     * @return {object}
     */
    Proto.singleton = function(key) {

        key = path + '-' + (key || true).toString();

        /**
         * Depending on where the method is called, the reference
         * to the object can be stored in another context
         * (other than kernel context)
         */
        if (typeof context._loaded === 'undefined') {
            context._loaded = [];
        }

        if (typeof context._loaded[key] !== 'undefined') {
            return context._loaded[key];
        }

        var obj      = Object.create(Proto.prototype || {});
        obj.kernel   = context.kernel || context;
        obj.instance = (context.kernel) ? context : null;

        Proto.call(obj, arguments[1], arguments[2], arguments[3], arguments[4], arguments[5]);

        context._loaded[key] = obj;
        return context._loaded[key];
    };

    return Proto;
};

/**
 * Generic replacement for console log with date
 *
 * @param {string|Error} message
 * @param {string} file
 * @param {Date} date
 * @return
 */
Kernel.prototype.log = function(message, file, date) {

    /**
     * Log the error stack if we got an Error object
     */
    if (message instanceof Error) {
        message = message.stack;
    }

    if (typeof file !== 'undefined') {

        var date     = this.utility.date(date, 'Y-m-d h:i:s')
            ,logfile = this.config.KERNEL_LOG_PATH
                     + (this.config.KERNEL_PROJECT_NAME || 'lemonade') + '.'
                     + file + '.log';

        fs.appendFile(
             logfile
            ,date + ' ' + message + '\n'
            ,function() {}
        );

    } else {
        util.log(message);
    }
};

/**
 * Export
 */
exports = module.exports = function(options) {
    new Kernel(
         options.config   || {}
        ,options.boot     || {}
        ,options.run      || {}
        ,options.shutdown || {}
    );
};